import types

import pkg_resources

from ..fixture import is_rule

_RULES_TAG = '__rules__'


def _autoload_plugin(plg):
    if isinstance(plg, types.ModuleType):
        if hasattr(plg, _RULES_TAG):
            return _autoload_plugin(getattr(plg, _RULES_TAG))
        else:
            raise ImportError(f'Not a module with rules inside - {repr(plg.__name__)}.')
    elif is_rule(plg) or isinstance(plg, (tuple, list)):
        return plg
    else:
        raise TypeError(f'Not a valid rule object, link or group - {repr(plg)}.')


POTC_PLUGIN_GROUP = 'potc_plugin'


def _iter_plugins():
    for ep in pkg_resources.iter_entry_points(group=POTC_PLUGIN_GROUP):
        yield ep.load()


def _load_plugins():
    return [_autoload_plugin(item) for item in _iter_plugins()]


def installed_loader():
    """
    Overview:
        Load rules from installed potc plugins.

    Returns:
        - rules: Rules from installed plugins.
    """
    return _load_plugins()



import os
import re
from abc import ABCMeta
from itertools import chain
from typing import Tuple, Mapping, Any, Type, Union

from isort import code as sort_code_imports
from yapf.yapflib.yapf_api import FormatCode

from .base import make_both_rules
from ..fixture import Addons, UnprocessableError
from ..fixture.imports import ImportStatement
from ..rules import installed_loader, system_all, default_all, native_all


class TranslationFailed(Exception):
    """
    Overview:
        Exception when translation is failed.
    """

    def __init__(self, obj):
        """
        Overview:
            Constructor of :class:`TranslationFailed`.

        Arguments:
            - obj: Object which is failed to be translated.

        """
        Exception.__init__(self, f'Translating failed for {repr(obj)}.')
        self.__obj = obj

    @property
    def obj(self):
        """
        Overview:
            The object which is failed to be translated.
        """
        return self.__obj


class ObjectTranslation:
    """
    Overview:
        Result of object translation.
    """

    def __init__(self, code: str, addon: Addons, trace: str):
        """
        Overview:
            Constructor of :class:`ObjectTranslation`.

        Arguments:
            - code (:obj:`str`): String translated code.
            - addon (:obj:`Addons`): Addon object for this translation.
            - trace (:obj:`str`): Rule name for the translated object.
        """
        self.__code = code
        self.__addon = addon
        self.__trace = trace

    @property
    def code(self) -> str:
        """
        Overview:
            String translated code.
        """
        return self.__code

    @property
    def addon(self) -> Addons:
        """
        Overview:
            Addon object for this translation.
        """
        return self.__addon

    @property
    def trace(self) -> str:
        """
        Overview:
            Rule name for the translated object.
        """
        return self.__trace

    @property
    def imports(self) -> Tuple[ImportStatement, ...]:
        """
        Overview:
            Import statements of this translation.
        """
        return self.addon.import_items

    def __str__(self):
        """
        Overview:
            Just get the string code, such as for the :func:`print` function.
        """
        return self.__code

    def __iter__(self):
        """
        Overview:
            Iterate this object.

        Examples::
            >>> ot = ObjectTranslation(cc, aa, tt)
            >>> c, a, t = ot
            >>> assert c is cc
            >>> assert a is aa
            >>> assert t is tt
        """
        yield self.__code
        yield self.__addon
        yield self.__trace


class VarsTranslation:
    """
    Overview:
        Result of vars translation.
    """

    def __init__(self, code: str, addons: Mapping[str, Addons]):
        """
        Overview:
            Constructor of :class:`VarsTranslation`.

        Arguments:
            - code (:obj:`str`): String translated code.
            - addons (:obj:`Mapping[str, Addons]`): Addon objects for the sub translations.
        """
        self.__code = code
        self.__addons = addons

    @property
    def code(self) -> str:
        """
        Overview:
            String translated code.
        """
        return self.__code

    @property
    def addons(self) -> Mapping[str, Addons]:
        """
        Overview:
            Addon objects for the sub translations.
        """
        return self.__addons

    def __str__(self) -> str:
        """
        Overview:
            Just get the string code, such as for the :func:`print` function.
        """
        return self.__code

    def __iter__(self):
        """
        Overview:
            Iterate this object.

        Examples::
            >>> ot = VarsTranslation(cc, aa)
            >>> c, a = ot
            >>> assert c is cc
            >>> assert a is aa
        """
        yield self.__code
        yield self.__addons


_VARS_KEY_PATTERN = re.compile(r'^[_a-zA-Z][_0-9a-zA-Z]*$')


class _AbstractTranslator(metaclass=ABCMeta):
    def __init__(self, rules):
        """
        Overview:
            Constructor of :class:`_AbstractTranslator`.

        Arguments:
            - rules: Rules for translator.
        """
        self.__rule, self.__plain_rule = make_both_rules(rules)

    def transobj(self, obj) -> ObjectTranslation:
        """
        Overview:
            Do translation to the given object ``obj``.

        Arguments:
            - obj: Object to be translated.

        Returns:
            - result (:obj:`ObjectTranslation`): Result of object translation.
        """
        addon = Addons(self.__plain_rule)
        try:
            _code, _trace_name = self.__rule(obj, addon)
            return ObjectTranslation(_code.strip(), addon, _trace_name)
        except UnprocessableError:
            raise TranslationFailed(obj)

    def __transvars(self, vars_: Mapping[str, Any]):
        codes, addons = [], []
        for key, value in sorted(vars_.items()):
            _result = self.transobj(value)
            _code, _addon = _result.code, _result.addon
            codes.append((key, _code))
            addons.append((key, _addon))

        _imports = sorted(set(chain(*map(lambda x: x[1].import_items, addons))), key=lambda x: x.key)
        return _imports, codes, addons

    def transvars(self, vars_: Mapping[str, Any], reformat=None, isort: bool = True) -> VarsTranslation:
        r"""
        Overview:
            Do translation to the given variables ``vars_``, and create a runnable code.

        Arguments:
            - vars\_ (:obj:`Mapping[str, Any]`): Variables to be translated.
            - reformat: Reformatter for the final source code, default is ``None`` which means \
                do not reformat the code.
            - isort (:obj:`bool`): Sort the import statements, default is ``True``.

        Returns:
            - result (:obj:`VarsTranslation`): Result of vars translation.
        """
        actual_vars = dict(vars_)
        for key in actual_vars.keys():
            if not _VARS_KEY_PATTERN.fullmatch(key):
                raise NameError(f'Invalid variable name - {repr(key)}.')

        actual_vars.update(dict(__all__=sorted(vars_.keys())))
        _import_lines, _assignments, _addons = self.__transvars(actual_vars)

        _assignment_lines = [f'{key} = {code}' for key, code in _assignments]
        _code = os.linesep.join(list(map(str, chain(_import_lines, _assignment_lines))))

        if isort:
            _code = sort_code_imports(_code)
        if reformat is not None:
            if callable(reformat):
                _code = reformat(_code)
            else:
                _code, _ = FormatCode(_code, style_config=reformat)
        return VarsTranslation(_code, dict(_addons))


class BaseTranslator(_AbstractTranslator):
    """
    Overview:
        Base translator for the common translation cases.

        This class is based on :class:`_AbstractTranslator`, contains only \
        system-leveled and default object-leveled translation rules.
    """

    def __init__(self, extend_rules=None):
        """
        Overview:
            Constructor of :class:`BaseTranslator`.

        Arguments:
            - extend_rules: Rules for base translator.
        """
        _AbstractTranslator.__init__(self, (system_all, extend_rules or [], default_all))


class BlankTranslator(BaseTranslator):
    """
    Overview:
        Blank translator for the common translation cases.

        This class is based on :class:`BaseTranslator`, contains all the native rules \
        in this project. May be the best choice for DIY and testings.

    .. note::

        The rules from the plugins will be not be here, \
        and will certainly not be applied when translation is processed.
    """

    def __init__(self, extend_rules=None):
        """
        Overview:
            Constructor of :class:`BlankTranslator`.

        Arguments:
            - extend_rules: Rules for base translator.
        """
        BaseTranslator.__init__(self, [(extend_rules or []), native_all])


class Translator(BlankTranslator):
    """
    Overview:
        Common translator for the common translation cases.

        This class is based on :class:`BlankTranslator`, contains all the \
        rules from the installed plugins. It is the best choice for being used directly.
    """

    def __init__(self, extend_rules=None):
        """
        Overview:
            Constructor of :class:`Translator`.

        Arguments:
            - extend_rules: Rules for base translator.
        """
        BlankTranslator.__init__(self, [(extend_rules or []), installed_loader()])


def autotrans(trans: Union[BaseTranslator, Type[BaseTranslator], list, tuple]) -> BaseTranslator:
    """
    Overview:
        Turn the given object ``trans`` to a :class:`BaseTranslator` instance.

    Arguments:
        - trans (:obj:`Union[BaseTranslator, Type[BaseTranslator], list, tuple]`): \
            Any object to be transformed to :class:`BaseTranslator` instance.

    Returns:
        - translator (:obj:`BaseTranslator`): Translator instance which can be directly used.

    """
    if isinstance(trans, BaseTranslator):
        return trans
    elif isinstance(trans, type) and issubclass(trans, BaseTranslator):
        return trans()
    else:
        return Translator(trans)



import builtins
import math
import types
from functools import partial
from typing import Type

from hbutils.reflection import get_callable_hint

from ...fixture import Addons, rule
from ...supports import function, raw_type, raw_object, typed_object
from ...supports.bin import dump_obj


@rule(type_=type(...))
def builtin_ellipsis():
    return '...'


@rule(type_=int)
def builtin_int(v: int):
    return repr(v)


@rule(type_=complex)
def builtin_complex(v: complex):
    return repr(v)


@rule(type_=float)
def builtin_float(v: float, addon: Addons):
    if math.isinf(v):
        return ('+' if v > 0 else '-') + str(addon.obj(math).inf)
    elif math.isnan(v):
        return addon.obj(math).nan
    else:
        if math.isclose(v, math.e):
            return addon.obj(math).e
        elif math.isclose(v, math.pi):
            return addon.obj(math).pi
        elif math.isclose(v, math.tau):
            return addon.obj(math).tau
        else:
            return repr(v)


@rule(type_=str)
def builtin_str(v: str):
    return repr(v)


@rule(type_=type(None))
def builtin_none():
    return 'None'


@rule(type_=range)
def builtin_range(v: range, addon: Addons):
    if v.step == 1:
        if v.start == 0:
            _args = (v.stop,)
        else:
            _args = (v.start, v.stop,)
    else:
        _args = (v.start, v.stop, v.step)
    return addon.val(range)(*_args)


@rule(type_=slice)
def builtin_slice(v: slice, addon: Addons):
    if v.step is None:
        if v.start is None:
            _args = (v.stop,)
        else:
            _args = (v.start, v.stop)
    else:
        _args = (v.start, v.stop, v.step)
    return addon.val(slice)(v.start, v.stop, v.step)


@rule(type_=list)
def builtin_list(v: list, addon: Addons):
    if type(v) == list:
        return f'[{", ".join(map(addon.rule, v))}]'
    else:
        return addon.obj(type(v))(list(v))


def _is_namedtuple_class(t: Type[tuple]) -> bool:
    _fields = getattr(t, '_fields', None)
    _field_ok = isinstance(_fields, tuple) and all(isinstance(name, str) for name in _fields)
    _asdict = getattr(t, '_asdict', None)
    _asdict_ok = isinstance(_asdict, types.FunctionType)
    return all((_field_ok, _asdict_ok))


@rule(type_=tuple)
def builtin_tuple(v: tuple, addon: Addons):
    if type(v) == tuple:
        return f'({", ".join(map(addon.rule, v))}{", " if len(v) == 1 else ""})'
    elif _is_namedtuple_class(type(v)):
        return addon.obj(type(v))(*tuple(v))
    else:
        return addon.obj(type(v))(tuple(v))


@rule(type_=(set, frozenset))
def builtin_set(v: set, addon: Addons):
    if type(v) == set:
        if len(v) > 0:
            return f'{{{", ".join(map(addon.rule, v))}}}'
        else:
            return 'set()'
    else:
        return addon.obj(type(v))(set(v))


@rule(type_=dict)
def builtin_dict(v: dict, addon: Addons):
    if type(v) == dict:
        return f'{{{", ".join(map(lambda e: f"{addon.rule(e[0])}: {addon.rule(e[1])}", v.items()))}}}'
    else:
        return addon.obj(type(v))(dict(v))


@rule(type_=(bytes, bytearray, memoryview))
def builtin_bytes(v: bytes, addon: Addons):
    if type(v) == bytes:
        return repr(v)
    else:
        return addon.val(type(v))(bytes(v))


_BUILTIN_ITEMS_NAMES = {id(value): key for key, value in builtins.__dict__.items() if not isinstance(value, type)}


@rule()
def builtin_items(v, addon: Addons):
    if id(v) in _BUILTIN_ITEMS_NAMES.keys():
        return _BUILTIN_ITEMS_NAMES[id(v)]
    else:
        addon.unprocessable()


# noinspection PyTypeChecker
@rule(type_=types.FunctionType)
def builtin_func(v, addon: Addons):
    return addon.obj(function)(v.__name__, get_callable_hint(v), dump_obj(v))


_TYPES_TYPE_NAMES = {getattr(types, name): name for name in filter(lambda x: x.endswith('Type'), dir(types))}
_BUILTIN_TYPE_NAMES = {value: key for key, value in builtins.__dict__.items() if isinstance(value, type)}


@rule(type_=type)
def builtin_type(v: type, addon: Addons):
    try:
        if v in _BUILTIN_TYPE_NAMES.keys():
            return _BUILTIN_TYPE_NAMES[v]
        elif v in _TYPES_TYPE_NAMES.keys():
            return getattr(addon.obj(types), _TYPES_TYPE_NAMES[v])
        else:
            return addon.obj(v)
    except (ImportError, TypeError):
        addon.unprocessable()


@rule(type_=type)
def builtin_raw_type(v: type, addon: Addons):
    _full_name = ((v.__module__ + '.') if hasattr(v, '__module__') else '') + v.__name__
    return addon.obj(raw_type)(_full_name, dump_obj(v))


@rule(type_=types.ModuleType)
def builtin_module(v, addon: Addons):
    return addon.obj(v)


@rule()
def builtin_object(v, addon: Addons):
    try:
        _i = addon.obj(type(v))
    except (ImportError, TypeError):
        _call = partial(addon.obj(raw_object))
    else:
        _call = partial(addon.obj(typed_object), type(v))

    return _call(dump_obj(v))


#: Overview:
#:      Basic builtin types.
#:
#: Items:
#:      - :class:`int`
#:      - :class:`complex`
#:      - :class:`float`
#:      - :class:`str`
#:      - :class:`none`
#:      - :class:`range`
#:      - :class:`slice`
#:      - :class:`ellipsis`
#:      - :class:`bytes`
builtin_basic = (
    builtin_int,
    builtin_complex,
    builtin_float,
    builtin_str,
    builtin_none,
    builtin_range,
    builtin_slice,
    builtin_ellipsis,
    builtin_bytes,
)

#: Overview:
#:      Builtin collection types
#:
#: Items:
#:      - :class:`list`
#:      - :class:`tuple`
#:      - :class:`set`
#:      - :class:`dict`
builtin_collection = (
    builtin_list,
    builtin_tuple,
    builtin_set,
    builtin_dict,
)

#: Overview:
#:      Builtin reflectance types
#:
#: Items:
#:      - Builtin items (items in module ``builtins``)
#:      - Functions
#:      - Types (including importable ones and unimportable ones)
#:      - Modules
builtin_reflect = (
    builtin_items,
    builtin_func,
    builtin_type,
    builtin_raw_type,
    builtin_module,
)

#: Overview:
#:      All Builtin types.
#:
builtin_all = [
    (
        builtin_basic,
        builtin_collection,
        builtin_reflect,
        builtin_object,
    ),
]



import copy
import io
from contextlib import contextmanager
from functools import lru_cache
from typing import Mapping, Any


@lru_cache()
def _transobj_func():
    from ..translate import transobj
    return transobj


@lru_cache()
def _transvars_func():
    from ..translate import transvars
    return transvars


def run_script(source: str) -> Mapping[str, Any]:
    vs = {key: value for key, value in globals().items() if key.startswith('__') and key.endswith('__')}
    exec('', vs)
    original_vs = copy.copy(vs)
    exec(source, vs)

    return {
        key: value for key, value in vs.items() if
        key not in original_vs.keys() or original_vs[key] is not vs[key]
    }


_TEST_OBJ = '_TEST_OBJ'


@contextmanager
def transobj_assert(obj: object, trans=None):
    """
    Overview:
        Assertion-based test for :func:`potc.translate.quick.transobj` function.

        This function is managed by ``contextmanager``, so you can use it with a ``with`` block.

    Arguments:
        - obj (:obj:`object`): Object to be translated
        - trans: Translator to be used. See :func:`potc.translate.quick.transobj`.

    Examples::
        >>> with transobj_assert([1, -2, 3.2, '1234']) as obj, name:
        ...     assert obj == [1, -2, 3.2, '1234']
        ...     assert name == 'builtin_list'

    .. note::

        When this function is called, the translated object code will be dumped into \
        a python file and then import from it, so please **make sure the code you created \
        can be supported by standard python grammar**.
    """
    _code, _addon, _name = _transobj_func()(obj, trans)

    with io.StringIO() as _source_file:
        for _import in sorted(set(_addon.import_items), key=lambda x: x.key):
            print(_import, file=_source_file)

        print(f'{_TEST_OBJ} = {_code}', file=_source_file)
        _source_code = _source_file.getvalue()

    _changes = run_script(_source_code)
    _obj = _changes[_TEST_OBJ]

    yield _obj, _name


@contextmanager
def transvars_assert(vars_: Mapping[str, Any], trans=None, reformat=None, isort: bool = True):
    r"""
    Overview:
        Assertion-based test for :func:`potc.translate.quick.transvars` function.

        This function is managed by ``contextmanager``, so you can use it with a ``with`` block.

    Arguments:
        - vars\_ (:obj:`Mapping[str, Any]`): The items to be translated, should be a mapping contains \
            all the values and their names.
        - trans: Translator to be used. See :func:`potc.translate.quick.transvars`.
        - reformat: Reformatter to be used. See :func:`potc.translate.quick.transvars`.
        - isort (:obj:`bool`): Sort the import statements. See :func:`potc.translate.quick.transvars`.

    Examples::
        >>> with transvars_assert({
        ...     'a': 233,
        ...     'b': [1, 2, 3, {'x': 1}],
        ... }) as (v, code):
        ...     assert v == {  # v is the mapping of the values
        ...         'a': 233,
        ...         'b': [1, 2, 3, {'x': 1}],
        ...     }
        ...     assert 'import' not in code  # code is the dumped source code

    .. note::

        When this function is called, the translated object code will be dumped into \
        a python file and then import from it, so please **make sure the code you created \
        can be supported by standard python grammar**.
    """
    _code, _addons = _transvars_func()(vars_, trans, reformat, isort)
    _changes = run_script(_code)

    assert '__all__' in _changes.keys(), f"No {repr('__all__')} variable found."
    _vars = {key: _changes[key] for key in _changes['__all__']}
    yield _vars, _code



from functools import reduce
from itertools import product
from operator import __or__
from typing import Union, Callable, List, Tuple, Optional

from hbutils.algorithm import topoids
from hbutils.reflection import dynamic_call

from .common import is_rule, rule_name, unprocessable, UnprocessableError


def build_chain(c: Union[List, Tuple, Callable],
                id_gen: Optional[Callable] = None) -> List[Callable]:
    """
    Overview:
        Build chain of the rules.
        Bases on orders which depends by tuples and lists.

    Arguments:
        - c (:obj:`Union[List, Tuple, Callable]`): Rule structure object, can be a list, \
            tuple or rule, or the nested structure of them.
        - id_gen (:obj:`Optional[Callable]`): Id generator of the rules, default is ``None`` \
            which means ``id`` function.

    Returns:
        - chain (:obj:`List[Callable]`): List of the rules in order.

    Example:
        >>> build_chain(r1)                        # [r1]
        >>> build_chain([(r1, r2), (r3, r2)])      # [r1, r3, r2] or [r3, r1, r2]
        >>> build_chain((r1, [r2, (r3, r4)], r5))  # [r1, r2, r3, r4, r5] or [r1, r3, r2, r4, r5] or [r1, r3, r4, r2, r5]
    """
    id_gen = id_gen or id
    rules = {}
    edges = set()

    def _rule_walk(ci: Union[List, Tuple, Callable]):
        if is_rule(ci):
            rules[id_gen(ci)] = ci
            return {id_gen(ci)}, {id_gen(ci)}, {id_gen(ci)}
        elif isinstance(ci, tuple):
            children = list(filter(lambda x: x[0], map(_rule_walk, ci)))
            if not children:
                return set(), set(), set()

            for (_, _, _lasts), (_, _firsts, _) in zip(children[:-1], children[1:]):
                for _last, _first in product(_lasts, _firsts):
                    edges.add((_last, _first))

            _all_ids = reduce(__or__, map(lambda x: x[0], children))
            _, _first_ids, _ = children[0]
            _, _, _last_ids = children[-1]
            return _all_ids, _first_ids, _last_ids
        elif isinstance(ci, list):
            children = list(filter(lambda x: x[0], map(_rule_walk, ci)))
            if not children:
                return set(), set(), set()

            _all_ids = reduce(__or__, map(lambda x: x[0], children))
            _first_ids = reduce(__or__, map(lambda x: x[1], children))
            _last_ids = reduce(__or__, map(lambda x: x[2], children))
            return _all_ids, _first_ids, _last_ids
        else:
            raise TypeError(f'Unknown element type in rule chain, '
                            f'list, tuple or decorated rule expected '
                            f'but {repr(type(ci).__name__)} found.')

    _rule_walk(c)

    all_ids = set(rules.keys())
    count = len(all_ids)
    _name_to_id = dict(map(lambda x: (rule_name(x), id_gen(x)), rules.values()))
    _names = sorted(set(map(lambda x: rule_name(x), rules.values())))

    if len(_names) < count:
        _duplicated_names = tuple(sorted(set(
            map(lambda x: x[0], filter(lambda x: x[0] == x[1], zip(_names[:-1], _names[1:])))
        )))
        raise NameError(f'Duplicate rule names in the chain - {repr(_duplicated_names)}.')

    _name_to_index = {name: nid for nid, name in enumerate(_names)}
    _id_to_index = {_name_to_id[name]: _name_to_index[name] for name in _names}
    _index_to_id = {_name_to_index[name]: _name_to_id[name] for name in _names}
    new_edges = set(map(lambda x: (_id_to_index[x[0]], _id_to_index[x[1]]), edges))

    try:
        order = topoids(count, new_edges)
    except ArithmeticError as err:
        _, conflict_indices = err.args
        _index_to_name = {nid: name for nid, name in enumerate(_names)}
        conflicted_names = tuple(sorted(map(lambda x: _index_to_name[x], conflict_indices)))
        raise ArithmeticError(
            f'Conflicted order with in {repr(conflicted_names)}.',
            tuple(map(lambda x: (x, rules[_index_to_id[_name_to_index[x]]]), conflicted_names)),
        )
    return [rules[_index_to_id[oid]] for oid in order]


def rules_combine(*rules):
    """
    Overview:
        Combined the ordered rule list to be a executable rule-like function (but not a rule object actually).

    Arguments:
        - rules: List of the rules (may be the result of function ``build_chain``.

    Returns:
        - func: Combined rule function.
    """
    for r in rules:
        if not is_rule(r):
            raise TypeError(f"Not a rule - {repr(r)}.")
    names = sorted(map(rule_name, rules))
    if len(set(names)) < len(names):
        duplicate_names = sorted(set(
            map(lambda p: p[0], filter(lambda p: p[0] == p[1], zip(names[:-1], names[1:])))
        ))
        raise KeyError(f'Duplicate names found in rule chain - {", ".join(map(repr, duplicate_names))}.')

    def _new_rule(v, addon):
        with addon.transaction():
            for _rule_item in map(dynamic_call, rules):
                try:
                    with addon.transaction():
                        _result = _rule_item(v, addon)
                except UnprocessableError:
                    continue
                else:
                    return str(_result), rule_name(_rule_item)

            unprocessable()

    return _new_rule



from ..fixture import Addons, AddonProxy, rule


@rule(type_=AddonProxy)
def system_addon_proxy(v: AddonProxy, addon: Addons):
    with addon.transaction():
        return str(v)


#: Overview:
#:      Rules for potc system.
#:
#: Items:
#:      - :class:`potc.fixture.addons.AddonProxy`.
system_all = [
    system_addon_proxy,
]



from .trans import autotrans, ObjectTranslation, VarsTranslation


def transobj(obj, trans=None) -> ObjectTranslation:
    """
    Overview:
        Process-oriented version of :meth:`potc.translate.trans.BaseTranslator.transobj`.

    Arguments:
        - obj: Object to be translated.
        - trans: Translator to be used, default is ``None`` which means use \
            an instance of :class:'potc.translate.trans.Translator'.

    Returns:
        - result (:obj:`ObjectTranslation`): Result of object translation.

    Examples::
        >>> import math
        >>> t = transobj([1, '2', 3.4, ..., None, math.e])
        >>> t
        <potc.translate.trans.ObjectTranslation object at 0x7f08a9520f10>
        >>> t.code
        "[1, '2', 3.4, ..., None, math.e]"
        >>> t.imports
        (<_SelfDirectImport 'import math'>,)
    """
    return autotrans(trans).transobj(obj)


def transvars(vars_, trans=None, reformat=None, isort: bool = True) -> VarsTranslation:
    """
    Overview:
        Process-oriented version of :meth:`potc.translate.trans.BaseTranslator.transvars`.

    Arguments:
        - vars\_ (:obj:`Mapping[str, Any]`): Variables to be translated.
        - trans: Translator to be used, default is ``None`` which means use \
            an instance of :class:'potc.translate.trans.Translator'.
        - reformat: Reformatter for the final source code, default is ``None`` which means \
            do not reformat the code.
        - isort (:obj:`bool`): Sort the import statements, default is ``True``.

    Returns:
        - result (:obj:`VarsTranslation`): Result of vars translation.

    Examples::
        >>> import math
        >>> t = transvars({
        ...     'a': [1, '2', 3.4, ..., None, math.e],
        ...     'b': dict(a='ds', b=(), c={3, 4, 5}),
        ... })
        >>> t
        <potc.translate.trans.VarsTranslation object at 0x7f08a95cd5b0>
        >>> print(t.code)
        import math
        __all__ = ['a', 'b']
        a = [1, '2', 3.4, ..., None, math.e]
        b = {'a': 'ds', 'b': (), 'c': {3, 4, 5}}
    """
    return autotrans(trans).transvars(vars_, reformat, isort)



from functools import wraps
from typing import Union, Tuple

import click
from hbutils.reflection import dynamic_call

from ...fixture.addons import Addons
from ...fixture.common import is_rule, rule, rule_name
from ...rules import builtin_all


def validator(func):
    func = dynamic_call(func)

    @wraps(func)
    def _new_func(ctx, param, value):
        return func(ctx=ctx, param=param, value=value)

    return _new_func


def multiple_validator(func):
    func = validator(func)

    @wraps(func)
    def _new_func(ctx, param, value):
        return [func(ctx, param, item) for item in value]

    return _new_func


_EXPECTED_TREE_ERRORS = (
    ValueError, TypeError, ImportError, AttributeError, ModuleNotFoundError,
    FileNotFoundError, IsADirectoryError, PermissionError, FileExistsError,
)

_EXCEPTION_WRAPPED = '__exception_wrapped__'


def err_validator(types: Union[type, Tuple[type, ...]]):
    def _decorator(func):
        func = validator(func)

        @wraps(func)
        def _new_func(ctx, param, value):
            try:
                return func(ctx, param, value)
            except click.BadParameter as err:
                raise err
            except types as err:
                _messages = [item for item in err.args if isinstance(item, str)]
                _final_message = _messages[0] if _messages else str(_messages)
                raise click.BadParameter(_final_message)

        return _new_func

    return _decorator


def _cli_builder(base_cli, *wrappers):
    _cli = None or base_cli
    for wrapper in wrappers:
        _cli = wrapper(_cli)
    return _cli


def _is_rule_rules(v):
    if is_rule(v):
        return True
    elif isinstance(v, (list, tuple)):
        return all(map(_is_rule_rules, v))
    else:
        return False


@rule()
def rules_rule(v, addon: Addons):
    if is_rule(v):
        return f'<{rule_name(v)}>'
    else:
        addon.unprocessable()


@rule(type_=list)
def rules_list(v, addon: Addons):
    if _is_rule_rules(v):
        return '[' + ', '.join(map(addon.rule, v)) + ']'
    else:
        addon.unprocessable()


@rule(type_=tuple)
def rule_tuple(v, addon: Addons):
    if _is_rule_rules(v):
        return '(' + ' --> '.join(map(addon.rule, v)) + ')'
    else:
        addon.unprocessable()


rules_struct = [
    (
        (rules_rule, rule_tuple, rules_list),
        builtin_all
    )
]



from contextlib import contextmanager
from functools import wraps
from unittest.mock import patch, MagicMock

import pkg_resources
from hbutils.reflection import quick_import_object

from ..rules.installed import POTC_PLUGIN_GROUP


class FakeEntryPoint:
    def __init__(self, name, *, dist=None, dist_name=None):
        self.name = name
        if not dist and dist_name:
            self.dist, _, _ = quick_import_object(dist_name)
        else:
            self.dist = dist

    def load(self):
        return self.dist


_UNTITLED_ID = -1


def _next_plugin_id():
    global _UNTITLED_ID
    _UNTITLED_ID += 1
    return f'_untitled_plugin_{_UNTITLED_ID}'
    pass


def _to_plugin(v) -> FakeEntryPoint:
    if isinstance(v, FakeEntryPoint):
        return v  # pragma: no cover
    elif isinstance(v, tuple):
        name, dist = v
        if isinstance(dist, str):
            return FakeEntryPoint(name, dist_name=dist)
        else:
            return FakeEntryPoint(name, dist=dist)
    else:
        return _to_plugin((_next_plugin_id(), v))


@contextmanager
def mock_potc_plugins(*plugins):
    """
    Overview:
        Mock potc plugins for unittest.

    :param plugins: String of plugin module, such as ``potc_dict.plugin``, which can be auto-imported.

    Examples::
        >>> from potc import transobj
        >>> from potc.testing import mock_potc_plugins
        >>>
        >>> # potc-dict is installed
        >>> print(transobj({'a': 1, 'b': [3, 'dfgk']}))
        dict(a=1, b=[3, 'dfgk'])
        >>>
        >>> with mock_potc_plugins():  # mock as no plugins
        ...     print(transobj({'a': 1, 'b': [3, 'dfgk']}))
        {'a': 1, 'b': [3, 'dfgk']}
        >>>
        >>> print(transobj({'a': 1, 'b': [3, 'dfgk']}))  # again
        dict(a=1, b=[3, 'dfgk'])
    """

    @wraps(pkg_resources.iter_entry_points)
    def _new_iter_func(group, name=None):
        if group == POTC_PLUGIN_GROUP:
            yield from filter(lambda x: (name is None or x.name == name), map(_to_plugin, plugins))
        else:
            yield from pkg_resources.iter_entry_points(group, name)  # pragma: no cover

    with patch('pkg_resources.iter_entry_points', MagicMock(side_effect=_new_iter_func)):
        yield



import importlib
import re
import types
import uuid
from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
from typing import Optional, Tuple


def try_import_info(obj, alias: Optional[str] = None) -> Tuple[str, ...]:
    """
    Overview:
        Try to get an import statement from given object.

    Arguments:
        - obj (:obj:`object`): Given object.
        - alias (:obj:`Optional[str]`): Alias name for import statement, \
            default is ``None`` which means do not alias.

    Returns:
        - import (:obj:`Tuple[str, ...]`): Import statement tuple.

    Example:
        >>> import dill
        >>> try_import_info(dill)  # ('import', 'dill')
        >>>
        >>> from potc import transvars
        >>> try_import_info(transvars)  # ('from', 'potc', 'import', 'transvars')
    """
    try:
        _simple = _raw_import_info(obj)
        _segments = _simple[1].split('.')
        if _simple[0] == 'import':
            combiner = lambda i: '.'.join([*_segments[:i], _segments[-1]])
            rg = range(0, len(_segments))
        else:
            combiner = lambda i: '.'.join(_segments[:i])
            rg = range(1, len(_segments) + 1)

        _simpler = lambda i: (_simple[0], combiner(i), *_simple[2:])
        for index in rg:
            _sp = _simpler(index)
            try:
                if _validate_import_info(*_sp) is obj:
                    return (*_sp, *(('as', alias) if alias else ()))
            except (TypeError, ImportError, AttributeError, ModuleNotFoundError):
                continue
        raise TypeError(f'Unable to import {repr(obj)}, validate fail.')
    except (TypeError, ImportError, AttributeError, ModuleNotFoundError):
        raise TypeError(f'Unable to import {repr(obj)}, error occurred when trying to trace.')


def _raw_import_info(obj):
    if isinstance(obj, types.ModuleType):
        _segments = obj.__name__.split('.')
        if len(_segments) > 1:
            return 'from', '.'.join(_segments[:-1]), 'import', _segments[-1]
        else:
            return 'import', obj.__name__
    elif hasattr(obj, '__module__') and hasattr(obj, '__name__'):
        return 'from', obj.__module__, 'import', obj.__name__
    else:
        raise TypeError(f'Unable to import {repr(obj)}.')


def _validate_import_info(*args):
    if args[0] == 'from':
        module = importlib.import_module(args[1])
        return getattr(module, args[3])
    elif args[0] == 'import':
        return importlib.import_module(args[1])


class ImportStatement(metaclass=ABCMeta):
    def __init__(self):
        """
        Overview:
            Constructor of :class:`ImportStatement`.
        """
        self.__uuid = uuid.uuid4()

    @property
    def uuid(self):
        """
        Overview:
            UUID of the import statement, will be unique when generated.
        """
        return self.__uuid

    @property
    @abstractmethod
    def target(self):
        """
        Overview:
            Import target.

            - ``from aa import bb``  --> bb.
            - ``from aa import bb as cc``  --> cc.
            - ``import aa``  --> aa.
            - ``import aa.bb``  --> bb.

            Just seen target as the actual sign name of the imported object.
        """
        raise NotImplementedError  # pragma: no cover

    @property
    @abstractmethod
    def key(self):
        """
        Overview:
            Key value of the import statement.
            Used for sorting.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _is_valid(self) -> bool:
        """
        Overview:
            Whether this import statement is valid.

        Return:
            - valid (:obj:`bool`): Is valid or not.
        """
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def _str(self) -> str:
        """
        Overview:
            String format of this import statement.

        Returns:
            - str (:obj:`str`): String format.
        """
        raise NotImplementedError  # pragma: no cover

    def _check_valid(self):
        """
        Overview:
            Check if this statement is valid.
            if not valid, raise ``ValueError``.
        """
        if not self._is_valid():
            raise ValueError(f'Invalid import statement - {repr(self._str())}.')

    @contextmanager
    def _with_check(self):
        """
        Overview:
            Open a context, do some update and make sure the final statement is valid.
            If not, this statement will be rolled back, and all the updates will be ignored.
        """
        _state = self.__getstate__()

        try:
            yield
            self._check_valid()
        except Exception as err:
            self.__setstate__(_state)
            raise err

    @abstractmethod
    def __getstate__(self):
        """
        Overview:
            Get value state of statement.

        Returns:
            - state: Value state.
        """
        return self.__uuid

    @abstractmethod
    def __setstate__(self, state):
        """
        Overview:
            Set value state of statement.
        """
        self.__uuid = state

    def __str__(self) -> str:
        """
        Overview:
            Get string format, will check its validity before return.

        Returns:
            - str (:obj:`str`): String format.
        """
        self._check_valid()
        return self._str()

    def __repr__(self) -> str:
        """
        Overview:
            Get representation format.

        Returns:
            - repr (:obj:`str`): Representation format.
        """
        return f'<{type(self).__name__} {repr(self._str())}>'


_SEGMENT_PATTERN = re.compile('[a-zA-Z0-9_]+')
_LINK_PATTERN = re.compile('[a-zA-Z0-9_]+(\\.[a-zA-Z0-9_]+)*')


class FromImport(ImportStatement):
    """
    Overview:
        ``from xxx import yyy as zzz`` formatted import statement.
    """

    def __init__(self, _from: str, _import: Optional[str] = None, _as: Optional[str] = None):
        """
        Overview:
            Constructor of ``FromImport``.

        Arguments:
            - \_from (:obj:`str`): From source.
            - \_import (:obj:`Optional[str]`): Import item, default is ``None``.
            - \_as (:obj:`Optional[str]`): Alias name, default is ``None`` which means no alias.
        """
        ImportStatement.__init__(self)
        self.__from = _from
        self.__import = _import
        self.__as = _as

    def import_(self, obj_name: str) -> 'FromImport':
        """
        Overview:
            Import from source.

        Arguments:
            - obj_name (:obj:`str`): Import object name.

        Returns:
            - self: Self object.
        """
        with self._with_check():
            self.__import = obj_name
            return self

    def as_(self, alias_name: Optional[str]) -> 'FromImport':
        """
        Overview:
            Alias imported item.

        Arguments:
            - alias_name (:obj:`Optional[str]`): Alias name.

        Returns:
            - self: Self object.
        """
        with self._with_check():
            self.__as = alias_name
            return self

    def _is_valid(self):
        if self.__from and self.__import:
            return (not self.__from or _LINK_PATTERN.fullmatch(self.__from)) and \
                   (not self.__import or _SEGMENT_PATTERN.fullmatch(self.__import)) and \
                   (not self.__as or _SEGMENT_PATTERN.fullmatch(self.__as))
        else:
            return False

    def _str(self):
        if not self.__as:
            return f'from {self.__from} import {self.__import}'
        else:
            return f'from {self.__from} import {self.__import} as {self.__as}'

    @property
    def target(self):
        self._check_valid()
        if self.__as:
            return self.__as
        else:
            return self.__import

    @property
    def key(self):
        self._check_valid()
        return f'{self.__from}.{self.__import}', 1, self.target

    def __hash__(self):
        return hash((self.__from, self.__import, self.__as))

    def __eq__(self, other):
        if other is self:
            return True
        elif type(other) == type(self):
            return (self.__from, self.__import, self.__as) == (other.__from, other.__import, other.__as)
        else:
            return False

    def __setstate__(self, state):
        self.__from, self.__import, self.__as, _state = state
        ImportStatement.__setstate__(self, _state)

    def __getstate__(self):
        return self.__from, self.__import, self.__as, ImportStatement.__getstate__(self)


class DirectImport(ImportStatement):
    """
    Overview:
        ``import xxx as yyy`` formatted import statement.
    """

    def __init__(self, _import: str, _as: Optional[str] = None):
        """
        Overview:
            Constructor of ``DirectImport``.

        Arguments:
            - \_import (:obj:`Optional[str]`): Import item, default is ``None``.
            - \_as (:obj:`Optional[str]`): Alias name, default is ``None`` which means no alias.
        """
        ImportStatement.__init__(self)
        self.__import = _import
        self.__as = _as

    def as_(self, alias_name: Optional[str]) -> 'DirectImport':
        """
        Overview:
            Alias imported item.

        Arguments:
            - alias_name (:obj:`Optional[str]`): Alias name.

        Returns:
            - self: Self object.
        """
        with self._with_check():
            self.__as = alias_name
            return self

    def _is_valid(self):
        if self.__import:
            return (not self.__import or _LINK_PATTERN.fullmatch(self.__import)) and \
                   (not self.__as or _SEGMENT_PATTERN.fullmatch(self.__as))
        else:
            return False

    def _str(self):
        if not self.__as:
            return f'import {self.__import}'
        else:
            return f'import {self.__import} as {self.__as}'

    @property
    def target(self):
        self._check_valid()
        if self.__as:
            return self.__as
        else:
            return self.__import.split('.')[-1]

    @property
    def key(self):
        self._check_valid()
        return f'{self.__import}', 0, self.target

    def __hash__(self):
        return hash((self.__import, self.__as))

    def __eq__(self, other):
        if other is self:
            return True
        elif type(other) == type(self):
            return (self.__import, self.__as) == (other.__import, other.__as)
        else:
            return False

    def __setstate__(self, state):
        self.__import, self.__as, _state = state
        ImportStatement.__setstate__(self, _state)

    def __getstate__(self):
        return self.__import, self.__as, ImportStatement.__getstate__(self)


class _SelfFormImport(FromImport):
    def __init__(self, _from: str, _import: Optional[str] = None, _as: Optional[str] = None, _this=None):
        FromImport.__init__(self, _from, _import, _as)
        self.__this = _this

    def import_(self, obj_name: str) -> '_SelfFormImport':
        _return = FromImport.import_(self, obj_name)
        self.__this.append(self)
        return self

    def as_(self, alias_name: Optional[str]) -> '_SelfFormImport':
        _return = FromImport.as_(self, alias_name)
        self.__this.append(self)
        return self

    def __getstate__(self):
        return FromImport.__getstate__(self), self.__this

    def __setstate__(self, state):
        _parent_state, self.__this = state
        FromImport.__setstate__(self, _parent_state)


class _SelfDirectImport(DirectImport):
    def __init__(self, _import: str, _as: Optional[str] = None, _this=None):
        DirectImport.__init__(self, _import, _as)
        self._check_valid()
        self.__this = _this
        self.__this.append(self)

    def as_(self, alias_name: Optional[str]) -> '_SelfDirectImport':
        _return = DirectImport.as_(self, alias_name)
        self.__this.append(self)
        return self

    def __getstate__(self):
        return DirectImport.__getstate__(self), self.__this

    def __setstate__(self, state):
        _parent_state, self.__this = state
        DirectImport.__setstate__(self, _parent_state)


class ImportPool:
    """
    Overview:
        Pool of import statements.
    """

    def __init__(self, *imports: ImportStatement):
        """
        Overview:
            Constructor of ``ImportPool``.

        Arguments:
            - imports: Import statements.
        """
        self.__imports = list(imports)
        self.__ids = {item.uuid: item.target for item in self.__imports}
        _this = self

    def append(self, import_: ImportStatement):
        """
        Overview:
            Append new statement.

        Arguments:
            - import\_: Import statement.
        """
        if import_.uuid not in self.__ids.keys():
            self.__imports.append(import_)
            self.__ids[import_.uuid] = import_.target

    def import_(self, _import: str) -> _SelfDirectImport:
        """
        Overview:
            Start a direct import statement.

        Arguments:
            - \_import (:obj:`str`): Import item.

        Returns:
            - statement (:obj:`DirectImport`): Direct import statement.
        """
        return _SelfDirectImport(_import, _this=self)

    def from_(self, _from: str) -> _SelfFormImport:
        """
        Overview:
            Start a from import statement.

        Arguments:
            - from (:obj:`str`): Import source.

        Returns:
            - statement (:obj:`FromImport`): From import statement.
        """
        return _SelfFormImport(_from, _this=self)

    @property
    def imports(self) -> Tuple['ImportStatement', ...]:
        """
        Overview:
            Get tuple of current import statements.

        Returns:
            - imports (:obj:`Tuple['ImportStatement', ...]`): Import statements.
        """
        return tuple(self.__imports)

    def __getstate__(self):
        """
        Overview:
            Get value state of import pool.

        Returns:
            - state: Value state.
        """
        return self.__imports, self.__ids

    def __setstate__(self, state):
        """
        Overview:
            Set value state of import pool.
        """
        self.__imports, self.__ids = state
